/**
 * test Constructor.extend() with natives
 *
 * YES!  You can extend native constructors, provided you are aware of a couple things.
 * In this set of tests, we'll use extend to subclass the Array constructor and provide workarounds for
 * for constructor initialization and [[Class]] dependent methods (toString()).
 *
 * BUT ~ 2 tests provided by @Raynos2 show that modifying the length property directly has side-
 * effects (like clobbering the array).
 *
 */

var test = require('tape');
var Constructor = require('../../constructor.js').Constructor;

/*
 * FIXTURE - fairly complete example of extending the native built-in Arrays.
 *
 * NOTE:  the first argument here could be the Array constructor.
 */
var SubArray = Constructor.extend([], {
  constructor: function () {
    [].slice.call(this);
    // stolen from kangax (http://bit.ly/W66vQW)
    this.push.apply(this, arguments);
  },
  
  // borrowed from kangax (http://bit.ly/W66vQW)
  last: function () {
    return this[this.length - 1];
  },

  /**
   * @method toString - must be defined explicitly
   *
   * kangax (http://bit.ly/W66vQW), David Herman in Effective JavaScript (item 40, pp. 106ff), and some 
   * learning tests tell us that the Array.prototype.toString() method fails on objects whose 
   * constructor is not really an Array - due to the internal use of [[Class]] by JavaScript engine - 
   * so we'll have to shim it, using an output string directly.
   *
   * NOTE - doesn't seem to work in IE 6, 7, or 8
   */
  toString: function () {
  
    var array = [];
    
    for (var i = 0; i < this.length; i += 1) {
        array.push(this[i]);
    }

    return array.join();
  },
  
  /**
   * @method concat - must be defined explicitly
   *
   * Array.concat works on the Array object and returns a new Array - we have to do the same thing, 
   * which is to return a new SubArray.  And that means we have to shim it.
   */
  concat: function () {
    
    var instance = new this.constructor();
    
    for (var i = 0; i < this.length; i += 1) {
        instance.push(this[i]);
    }
    for (i = 0; i < arguments.length; i += 1) {
        instance.push(arguments[i]);
    }
    
    return instance;
  } // watch out for trailing commas in IE 6-8!!!
});

  
test('SubArray - verify the toString() fix', function (t) {

  var b = new SubArray('first', 'middle', 'last');
  
  t.strictEqual(b.toString(), 'first,middle,last', 'toString() fix');
  t.end();
});

test('SubArray - verify the concat fix', function (t) {

  var b = new SubArray('first');
  
  b = b.concat('last');
  
  t.strictEqual(b.last(), 'last');
  t.end();
});

test('SubArray - verify custom method - last', function (t) {

  var b = new SubArray('first', 'middle', 'last');
  
  t.strictEqual(b.last(), 'last');
  t.end();
});

test('SubArray - index access', function (t) {

  var b = new SubArray('first', 'middle', 'last');
  
  t.strictEqual(b[0], 'first');
  t.strictEqual(b[1], 'middle');
  t.strictEqual(b[2], 'last');
  t.end();
});

// failing tests offered by @Raynos2 ~ converted to "passing" by expecting exceptions
test('SubArray - length append', function (t) {

  var message = 'length not modified';
  var b = new SubArray(0, 1);
  
  var len = b.length;
  
  b[len] = len;
  
  // fails
  //t.strictEqual(b.length, len + 1);

  try {   
    if (b.length === len) {
      throw new Error(message);
    }
    t.fail('length unexpectedly modified');
  } catch (e) {
    t.strictEqual(e.message, message);
  }
  
  t.end();
});

// failing tests offered by @Raynos2 ~ converted to "passing" by expecting exceptions
test('SubArray - length reset', function (t) {

  var b = new SubArray(0, 1);
  
  b.length = 0;
  
  try {
    if ("0" in b) {
      throw new Error('zero not removed');
    }
    t.fail('resetting length somehow removed zero');
  } catch (e) {
    t.pass(e.message);
  }
  
  try {
    if (typeof b[0] !== 'undefined') {
      throw new Error('zero not undefined');
    }
    t.fail('resetting length somehow undefined zero');
  } catch (e) {
    t.pass(e.message);
  }

  t.end();
});

test('SubArray - length after create', function (t) {

  var b = new SubArray('first', 'middle', 'last');

  t.strictEqual(b.length, 3);
  t.end();
});

test('SubArray - reverse', function (t) {

  var b = new SubArray('first', 'middle', 'last');
  
  t.strictEqual(b.reverse().last(), 'first');
  t.end();
});

test('SubArray - sort', function (t) {

  var b = new SubArray('first', 'middle', 'last');
  
  t.strictEqual(b.sort().last(), 'middle');
  t.end();
});

test('SubArray - shift', function (t) {

  var b = new SubArray('first', 'middle', 'last');
  
  b.shift();
  
  t.strictEqual(b[0], 'middle');
  t.strictEqual(b.length, 2);
  t.end();
});

test('SubArray - pop', function (t) {

  var b = new SubArray('first', 'middle', 'last');
  
  t.strictEqual(b.pop(), 'last');
  t.end();
});

test('SubArray - push', function (t) {

  var b = new SubArray('first');
  
  b.push('last');
  
  t.strictEqual(b.last(), 'last');
  t.strictEqual(b.length, 2);
  t.end();
});

test('SubArray - splice', function (t) {

  var b = new SubArray('first');
  
  b.splice(1, 1, 'last');
  
  t.strictEqual(b.last(), 'last');
  t.end();
});

test('SubArray - unshift', function (t) {

  var b = new SubArray('first');
  
  b.unshift('last');
  
  t.strictEqual(b.last(), 'first');
  t.end();
});


/* template */
/*
test('$1 should return $2', function (t) {
  t.plan(1);
  t.fail('next test to implement');
});
*/
